@@ -17,7 +17,7 @@ import (
 
 	"github.com/BurntSushi/toml"
 	"github.com/containers/common/pkg/hooks"
-	conmonconfig "github.com/containers/conmon/runner/config"
+	conmonrsClient "github.com/containers/conmon-rs/pkg/client"
 	"github.com/containers/image/v5/pkg/sysregistriesv2"
 	"github.com/containers/image/v5/types"
 	"github.com/containers/storage"
@@ -27,6 +27,7 @@ import (
 	selinux "github.com/opencontainers/selinux/go-selinux"
 	"github.com/sirupsen/logrus"
 	"k8s.io/utils/cpuset"
+	"k8s.io/utils/ptr"
 	"tags.cncf.io/container-device-interface/pkg/cdi"
 
 	"github.com/cri-o/cri-o/internal/config/apparmor"
@@ -45,7 +46,7 @@ import (
 	"github.com/cri-o/cri-o/internal/log"
 	"github.com/cri-o/cri-o/internal/storage/references"
 	"github.com/cri-o/cri-o/pkg/annotations"
-	"github.com/cri-o/cri-o/server/otel-collector/collectors"
+	"github.com/cri-o/cri-o/server/metrics/collectors"
 	"github.com/cri-o/cri-o/server/useragent"
 	"github.com/cri-o/cri-o/utils"
 	"github.com/cri-o/cri-o/utils/cmdrunner"
@@ -272,8 +273,41 @@ type RuntimeHandler struct {
 	// Default annotations specified for runtime handler if they're not overridden by
 	// the pod spec.
 	DefaultAnnotations map[string]string `toml:"default_annotations,omitempty"`
+
+	// StreamWebsockets can be used to enable the WebSocket protocol for
+	// container exec, attach and port forward.
+	//
+	// conmon-rs (runtime_type = "pod") supports this configuration for exec
+	// and attach. Forwarding ports will be supported in future releases.
+	StreamWebsockets bool `toml:"stream_websockets,omitempty"`
+
+	// ExecCPUAffinity specifies which CPU is used when exec-ing the container.
+	// The valid values are:
+	// "":
+	//   Use runtime default.
+	// "first":
+	//   When it has only exclusive cpuset, use the first CPU in the exclusive cpuset.
+	//   When it has both shared and exclusive cpusets, use first CPU in the shared cpuset.
+	ExecCPUAffinity ExecCPUAffinityType `toml:"exec_cpu_affinity,omitempty"`
+
+	// SeccompProfile is the absolute path of the seccomp.json profile which is used as the
+	// default for the runtime. This configuration takes precedence over runtime config seccomp_profile.
+	// If set to "", the runtime config seccomp_profile will be used.
+	// If that is also set to "", the internal default seccomp profile will be applied.
+	SeccompProfile string `toml:"seccomp_profile,omitempty"`
+
+	// seccompConfig is the seccomp configuration for the handler.
+	seccompConfig *seccomp.Config
 }
 
+type ExecCPUAffinityType string
+
+const (
+	ExecCPUAffinityTypeDefault   ExecCPUAffinityType = ""
+	ExecCPUAffinityTypeFirst     ExecCPUAffinityType = "first"
+	runtimeSeccompProfileDefault string              = ""
+)
+
 // Multiple runtime Handlers in a map.
 type Runtimes map[string]*RuntimeHandler
 
@@ -356,8 +390,13 @@ type RuntimeConfig struct {
 
 	// SeccompProfile is the seccomp.json profile path which is used as the
 	// default for the runtime.
+	// If set to "" or not found, the internal default seccomp profile will be used.
 	SeccompProfile string `toml:"seccomp_profile"`
 
+	// PrivilegedSeccompProfile can be set to enable a seccomp profile for
+	// privileged containers from the local path.
+	PrivilegedSeccompProfile string `toml:"privileged_seccomp_profile"`
+
 	// ApparmorProfile is the apparmor profile name which is used as the
 	// default for the runtime.
 	ApparmorProfile string `toml:"apparmor_profile"`
@@ -561,6 +600,7 @@ type ImageConfig struct {
 	SignaturePolicyDir string `toml:"signature_policy_dir"`
 	// InsecureRegistries is a list of registries that must be contacted w/o
 	// TLS verification.
+	// Deprecated: use `insecure` in `registries.conf` instead.
 	InsecureRegistries []string `toml:"insecure_registries"`
 	// ImageVolumes controls how volumes specified in image config are handled
 	ImageVolumes ImageVolumesType `toml:"image_volumes"`
@@ -575,6 +615,13 @@ type ImageConfig struct {
 	// calculating the pull progress interval to pullProgressTimeout / 10.
 	// Can be set to 0 to disable the timeout as well as the progress output.
 	PullProgressTimeout time.Duration `toml:"pull_progress_timeout"`
+	// OCIArtifactMountSupport is used to determine if CRI-O should support OCI Artifacts.
+	OCIArtifactMountSupport bool `toml:"oci_artifact_mount_support"`
+	// ShortNameMode describes the mode of short name resolution.
+	// The valid values are "enforcing" and "disabled".
+	// If "enforcing", an image pull will fail if a short name is used, but the results are ambiguous.
+	// If "disabled", the first result will be chosen.
+	ShortNameMode string `toml:"short_name_mode"`
 }
 
 // NetworkConfig represents the "crio.network" TOML config table.
@@ -704,7 +751,12 @@ type tomlConfig struct {
 
 // SetSystemContext configures the SystemContext used by containers/image library.
 func (t *tomlConfig) SetSystemContext(c *Config) {
-	c.SystemContext.BigFilesTemporaryDir = c.ImageConfig.BigFilesTemporaryDir
+	c.SystemContext.BigFilesTemporaryDir = c.BigFilesTemporaryDir
+	c.SystemContext.ShortNameMode = ptr.To(types.ShortNameModeEnforcing)
+
+	if c.ShortNameMode == "disabled" {
+		c.SystemContext.ShortNameMode = ptr.To(types.ShortNameModeDisabled)
+	}
 }
 
 func (t *tomlConfig) toConfig(c *Config) {
@@ -742,10 +794,13 @@ const configLogPrefix = "Updating config from "
 // otherwise.
 func (c *Config) UpdateFromFile(ctx context.Context, path string) error {
 	log.Infof(ctx, configLogPrefix+"single file: %s", path)
+
 	if err := c.UpdateFromDropInFile(ctx, path); err != nil {
 		return fmt.Errorf("update config from drop-in file: %w", err)
 	}
+
 	c.singleConfigPath = path
+
 	return nil
 }
 
@@ -758,6 +813,7 @@ func (c *Config) UpdateFromDropInFile(ctx context.Context, path string) error {
 	log.Infof(ctx, configLogPrefix+"drop-in file: %s", path)
 	// keeps the storage options from storage.conf and merge it to crio config
 	var storageOpts []string
+
 	storageOpts = append(storageOpts, c.StorageOptions...)
 	// storage configurations from storage.conf, if crio config has no values for these, they will be merged to crio config
 	graphRoot := c.Root
@@ -777,21 +833,24 @@ func (c *Config) UpdateFromDropInFile(ctx context.Context, path string) error {
 		return fmt.Errorf("unable to decode configuration %v: %w", path, err)
 	}
 
-	storageOpts = append(storageOpts, t.Crio.RootConfig.StorageOptions...)
+	storageOpts = append(storageOpts, t.Crio.StorageOptions...)
 	storageOpts = removeDupStorageOpts(storageOpts)
-	t.Crio.RootConfig.StorageOptions = storageOpts
+	t.Crio.StorageOptions = storageOpts
 	// inherits storage configurations from storage.conf
 	if t.Crio.Root == "" {
 		t.Crio.Root = graphRoot
 	}
+
 	if t.Crio.RunRoot == "" {
 		t.Crio.RunRoot = runRoot
 	}
+
 	if t.Crio.Storage == "" {
 		t.Crio.Storage = storageDriver
 	}
 
 	t.toConfig(c)
+
 	return nil
 }
 
@@ -799,17 +858,22 @@ func (c *Config) UpdateFromDropInFile(ctx context.Context, path string) error {
 // keeps the last appearance.
 func removeDupStorageOpts(storageOpts []string) []string {
 	var resOpts []string
+
 	opts := make(map[string]bool)
 	for i := len(storageOpts) - 1; i >= 0; i-- {
 		if ok := opts[storageOpts[i]]; ok {
 			continue
 		}
+
 		opts[storageOpts[i]] = true
+
 		resOpts = append(resOpts, storageOpts[i])
 	}
+
 	for i, j := 0, len(resOpts)-1; i < j; i, j = i+1, j-1 {
 		resOpts[i], resOpts[j] = resOpts[j], resOpts[i]
 	}
+
 	return resOpts
 }
 
@@ -817,9 +881,11 @@ func removeDupStorageOpts(storageOpts []string) []string {
 // configuration for it.
 func (c *Config) UpdateFromPath(ctx context.Context, path string) error {
 	log.Infof(ctx, configLogPrefix+"path: %s", path)
+
 	if _, err := os.Stat(path); err != nil && os.IsNotExist(err) {
 		return nil
 	}
+
 	if err := filepath.Walk(path,
 		func(p string, info os.FileInfo, err error) error {
 			if err != nil {
@@ -828,11 +894,14 @@ func (c *Config) UpdateFromPath(ctx context.Context, path string) error {
 			if info.IsDir() {
 				return nil
 			}
+
 			return c.UpdateFromDropInFile(ctx, p)
 		}); err != nil {
 		return fmt.Errorf("walk path: %w", err)
 	}
+
 	c.dropInConfigDir = path
+
 	return nil
 }
 
@@ -862,6 +931,7 @@ func (c *Config) ToString() (string, error) {
 // fails, which should never happen at all because of general type safeness.
 func (c *Config) ToBytes() ([]byte, error) {
 	var buffer bytes.Buffer
+
 	e := toml.NewEncoder(&buffer)
 
 	tc := tomlConfig{}
@@ -880,11 +950,14 @@ func DefaultConfig() (*Config, error) {
 	if err != nil {
 		return nil, err
 	}
+
 	cgroupManager := cgmgr.New()
+
 	ua, err := useragent.Get()
 	if err != nil {
 		return nil, fmt.Errorf("get user agent: %w", err)
 	}
+
 	return &Config{
 		Comment: "# ",
 		SystemContext: &types.SystemContext{
@@ -910,53 +983,16 @@ func DefaultConfig() (*Config, error) {
 			GRPCMaxSendMsgSize: defaultGRPCMaxMsgSize,
 			GRPCMaxRecvMsgSize: defaultGRPCMaxMsgSize,
 		},
-		RuntimeConfig: RuntimeConfig{
-			AllowedDevices:     []string{"/dev/fuse", "/dev/net/tun"},
-			DecryptionKeysPath: "/etc/crio/keys/",
-			DefaultRuntime:     DefaultRuntime,
-			Runtimes: Runtimes{
-				DefaultRuntime: defaultRuntimeHandler(),
-			},
-			SELinux:                     selinuxEnabled(),
-			ApparmorProfile:             apparmor.DefaultProfile,
-			BlockIOConfigFile:           DefaultBlockIOConfigFile,
-			BlockIOReload:               DefaultBlockIOReload,
-			IrqBalanceConfigFile:        DefaultIrqBalanceConfigFile,
-			RdtConfigFile:               rdt.DefaultRdtConfigFile,
-			CgroupManagerName:           cgroupManager.Name(),
-			PidsLimit:                   DefaultPidsLimit,
-			ContainerExitsDir:           containerExitsDir,
-			ContainerAttachSocketDir:    conmonconfig.ContainerAttachSocketDir,
-			MinimumMappableUID:          -1,
-			MinimumMappableGID:          -1,
-			LogSizeMax:                  DefaultLogSizeMax,
-			CtrStopTimeout:              defaultCtrStopTimeout,
-			DefaultCapabilities:         capabilities.Default(),
-			LogLevel:                    "info",
-			HooksDir:                    []string{hooks.DefaultDir},
-			CDISpecDirs:                 cdi.DefaultSpecDirs,
-			NamespacesDir:               defaultNamespacesDir,
-			DropInfraCtr:                true,
-			IrqBalanceConfigRestoreFile: DefaultIrqBalanceConfigRestoreFile,
-			seccompConfig:               seccomp.New(),
-			apparmorConfig:              apparmor.New(),
-			blockioConfig:               blockio.New(),
-			cgroupManager:               cgroupManager,
-			deviceConfig:                device.New(),
-			namespaceManager:            nsmgr.New(defaultNamespacesDir, ""),
-			rdtConfig:                   rdt.New(),
-			ulimitsConfig:               ulimits.New(),
-			HostNetworkDisableSELinux:   true,
-			DisableHostPortMapping:      false,
-			EnableCriuSupport:           true,
-		},
+		RuntimeConfig: *DefaultRuntimeConfig(cgroupManager),
 		ImageConfig: ImageConfig{
-			DefaultTransport:    "docker://",
-			PauseImage:          DefaultPauseImage,
-			PauseCommand:        "/pause",
-			ImageVolumes:        ImageVolumesMkdir,
-			SignaturePolicyDir:  "/etc/crio/policies",
-			PullProgressTimeout: 10 * time.Second,
+			DefaultTransport:        "docker://",
+			PauseImage:              DefaultPauseImage,
+			PauseCommand:            "/pause",
+			ImageVolumes:            ImageVolumesMkdir,
+			SignaturePolicyDir:      "/etc/crio/policies",
+			PullProgressTimeout:     0,
+			OCIArtifactMountSupport: true,
+			ShortNameMode:           "enforcing",
 		},
 		NetworkConfig: NetworkConfig{
 			NetworkDir: cniConfigDir,
@@ -976,6 +1012,50 @@ func DefaultConfig() (*Config, error) {
 	}, nil
 }
 
+// DefaultRuntimeConfig returns the default Runtime configs.
+func DefaultRuntimeConfig(cgroupManager cgmgr.CgroupManager) *RuntimeConfig {
+	return &RuntimeConfig{
+		AllowedDevices:     []string{"/dev/fuse", "/dev/net/tun"},
+		DecryptionKeysPath: "/etc/crio/keys/",
+		DefaultRuntime:     DefaultRuntime,
+		Runtimes: Runtimes{
+			DefaultRuntime: defaultRuntimeHandler(cgroupManager.IsSystemd()),
+		},
+		SELinux:                     selinuxEnabled(),
+		ApparmorProfile:             apparmor.DefaultProfile,
+		BlockIOConfigFile:           DefaultBlockIOConfigFile,
+		BlockIOReload:               DefaultBlockIOReload,
+		IrqBalanceConfigFile:        DefaultIrqBalanceConfigFile,
+		RdtConfigFile:               rdt.DefaultRdtConfigFile,
+		CgroupManagerName:           cgroupManager.Name(),
+		PidsLimit:                   DefaultPidsLimit,
+		ContainerExitsDir:           containerExitsDir,
+		ContainerAttachSocketDir:    ContainerAttachSocketDir,
+		MinimumMappableUID:          -1,
+		MinimumMappableGID:          -1,
+		LogSizeMax:                  DefaultLogSizeMax,
+		CtrStopTimeout:              defaultCtrStopTimeout,
+		DefaultCapabilities:         capabilities.Default(),
+		LogLevel:                    "info",
+		HooksDir:                    []string{hooks.DefaultDir},
+		CDISpecDirs:                 cdi.DefaultSpecDirs,
+		NamespacesDir:               defaultNamespacesDir,
+		DropInfraCtr:                true,
+		IrqBalanceConfigRestoreFile: DefaultIrqBalanceConfigRestoreFile,
+		seccompConfig:               seccomp.New(),
+		apparmorConfig:              apparmor.New(),
+		blockioConfig:               blockio.New(),
+		cgroupManager:               cgroupManager,
+		deviceConfig:                device.New(),
+		namespaceManager:            nsmgr.New(defaultNamespacesDir, ""),
+		rdtConfig:                   rdt.New(),
+		ulimitsConfig:               ulimits.New(),
+		HostNetworkDisableSELinux:   true,
+		DisableHostPortMapping:      false,
+		EnableCriuSupport:           true,
+	}
+}
+
 // Validate is the main entry point for library configuration validation.
 // The parameter `onExecution` specifies if the validation should include
 // execution checks. It returns an `error` on validation failure, otherwise
@@ -1003,10 +1083,18 @@ func (c *Config) Validate(onExecution bool) error {
 		return fmt.Errorf("validating runtime config: %w", err)
 	}
 
-	c.RuntimeConfig.seccompConfig.SetNotifierPath(
+	c.seccompConfig.SetNotifierPath(
 		filepath.Join(filepath.Dir(c.Listen), "seccomp"),
 	)
 
+	for name := range c.Runtimes {
+		if c.Runtimes[name].seccompConfig != nil {
+			c.Runtimes[name].seccompConfig.SetNotifierPath(
+				filepath.Join(filepath.Dir(c.Listen), "seccomp"),
+			)
+		}
+	}
+
 	if err := c.ImageConfig.Validate(onExecution); err != nil {
 		return fmt.Errorf("validating image config: %w", err)
 	}
@@ -1038,10 +1126,21 @@ func (c *APIConfig) Validate(onExecution bool) error {
 	if c.GRPCMaxSendMsgSize <= 0 {
 		c.GRPCMaxSendMsgSize = defaultGRPCMaxMsgSize
 	}
+
 	if c.GRPCMaxRecvMsgSize <= 0 {
 		c.GRPCMaxRecvMsgSize = defaultGRPCMaxMsgSize
 	}
 
+	if c.StreamEnableTLS {
+		if c.StreamTLSCert == "" {
+			return errors.New("stream TLS cert path is empty")
+		}
+
+		if c.StreamTLSKey == "" {
+			return errors.New("stream TLS key path is empty")
+		}
+	}
+
 	if onExecution {
 		return RemoveUnusedSocket(c.Listen)
 	}
@@ -1079,9 +1178,11 @@ func (c *RootConfig) Validate(onExecution bool) error {
 		if !filepath.IsAbs(c.LogDir) {
 			return errors.New("log_dir is not an absolute path")
 		}
+
 		if err := os.MkdirAll(c.LogDir, 0o700); err != nil {
 			return fmt.Errorf("invalid log_dir: %w", err)
 		}
+
 		store, err := c.GetStore()
 		if err != nil {
 			return fmt.Errorf("failed to get store to set defaults: %w", err)
@@ -1158,6 +1259,7 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 		if err != nil {
 			return fmt.Errorf("%q not found in $PATH: %w", tasksetBinary, err)
 		}
+
 		cmdrunner.PrependCommandsWith(executable, "--cpu-list", set.String())
 	}
 
@@ -1172,6 +1274,7 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 		if err != nil {
 			return fmt.Errorf("unable to update cgroup manager: %w", err)
 		}
+
 		c.cgroupManager = cgroupManager
 
 		if err := c.ValidateRuntimes(); err != nil {
@@ -1188,21 +1291,28 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 		// it does not exist but can be created
 		// otherwise, we skip
 		hooksDirs := []string{}
+
 		for _, hooksDir := range c.HooksDir {
 			if err := utils.IsDirectory(hooksDir); err != nil {
 				if !os.IsNotExist(err) {
 					logrus.Warnf("Skipping invalid hooks directory: %s exists but is not a directory", hooksDir)
+
 					continue
 				}
+
 				if err := os.MkdirAll(hooksDir, 0o755); err != nil {
 					logrus.Debugf("Failed to create requested hooks dir: %v", err)
+
 					continue
 				}
 			}
+
 			logrus.Debugf("Using hooks directory: %s", hooksDir)
 			hooksDirs = append(hooksDirs, hooksDir)
+
 			continue
 		}
+
 		c.HooksDir = hooksDirs
 
 		if err := cdi.Configure(cdi.WithSpecDirs(c.CDISpecDirs...)); err != nil {
@@ -1222,6 +1332,7 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 		if c.EnableCriuSupport {
 			if err := validateCriuInPath(); err != nil {
 				c.EnableCriuSupport = false
+
 				logrus.Infof("Checkpoint/restore support disabled: CRIU binary not found int $PATH")
 			} else {
 				logrus.Infof("Checkpoint/restore support enabled")
@@ -1230,14 +1341,20 @@ func (c *RuntimeConfig) Validate(systemContext *types.SystemContext, onExecution
 			logrus.Infof("Checkpoint/restore support disabled via configuration")
 		}
 
-		if err := c.seccompConfig.LoadProfile(c.SeccompProfile); err != nil {
+		if c.SeccompProfile == "" {
+			if err := c.seccompConfig.LoadDefaultProfile(); err != nil {
+				return fmt.Errorf("unable to load default seccomp profile: %w", err)
+			}
+		} else if err := c.seccompConfig.LoadProfile(c.SeccompProfile); err != nil {
 			if !errors.Is(err, os.ErrNotExist) {
 				return fmt.Errorf("unable to load seccomp profile: %w", err)
 			}
 
-			logrus.Info("Specified profile does not exist on disk")
+			// Fallback to the internal default in order not to break upgrade paths.
+			logrus.Info("Seccomp profile does not exist on disk, fallback to internal default profile")
+
 			if err := c.seccompConfig.LoadDefaultProfile(); err != nil {
-				return fmt.Errorf("load default seccomp profile: %w", err)
+				return fmt.Errorf("unable to load default seccomp profile: %w", err)
 			}
 		}
 
@@ -1280,7 +1397,7 @@ func (c *RuntimeConfig) ValidateDefaultRuntime() error {
 	// The default config sets crun and its path in the runtimes map, so check for that
 	// first. If it does not exist then we add runc + its path to the runtimes map.
 	if _, ok := c.Runtimes[DefaultRuntime]; !ok {
-		c.Runtimes[DefaultRuntime] = defaultRuntimeHandler()
+		c.Runtimes[DefaultRuntime] = defaultRuntimeHandler(c.cgroupManager.IsSystemd())
 	}
 	// Set the DefaultRuntime to runc so we don't fail further along in the code
 	c.DefaultRuntime = DefaultRuntime
@@ -1288,7 +1405,18 @@ func (c *RuntimeConfig) ValidateDefaultRuntime() error {
 	return nil
 }
 
-func defaultRuntimeHandler() *RuntimeHandler {
+// getDefaultMonitorGroup checks which defaultmonitor group to use
+// for cgroupfs it is empty.
+func getDefaultMonitorGroup(isSystemd bool) string {
+	monitorGroup := ""
+	if isSystemd {
+		monitorGroup = defaultMonitorCgroup
+	}
+
+	return monitorGroup
+}
+
+func defaultRuntimeHandler(isSystemd bool) *RuntimeHandler {
 	return &RuntimeHandler{
 		RuntimeType: DefaultRuntimeType,
 		RuntimeRoot: DefaultRuntimeRoot,
@@ -1300,7 +1428,9 @@ func defaultRuntimeHandler() *RuntimeHandler {
 			"PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
 		},
 		ContainerMinMemory: units.BytesSize(defaultContainerMinMemoryCrun),
-		MonitorCgroup:      defaultMonitorCgroup,
+		MonitorCgroup:      getDefaultMonitorGroup(isSystemd),
+		ExecCPUAffinity:    ExecCPUAffinityTypeDefault,
+		SeccompProfile:     runtimeSeccompProfileDefault,
 	}
 }
 
@@ -1347,6 +1477,7 @@ func (c *RuntimeConfig) ValidateRuntimes() error {
 	for _, invalidHandlerName := range failedValidation {
 		delete(c.Runtimes, invalidHandlerName)
 	}
+
 	c.initializeRuntimeFeatures()
 
 	return nil
@@ -1357,6 +1488,7 @@ func (c *RuntimeConfig) initializeRuntimeFeatures() {
 		versionOutput, err := cmdrunner.CombinedOutput(handler.RuntimePath, "--version")
 		if err != nil {
 			logrus.Errorf("Unable to determine version of runtime handler %q: %v", name, err)
+
 			continue
 		}
 
@@ -1368,12 +1500,14 @@ func (c *RuntimeConfig) initializeRuntimeFeatures() {
 		output, err := cmdrunner.CombinedOutput(handler.RuntimePath, "features")
 		if err != nil {
 			logrus.Errorf("Getting %s OCI runtime features failed: %s: %v", handler.RuntimePath, output, err)
+
 			continue
 		}
 
 		// Ignore error if we can't load runtime features.
 		if err := handler.LoadRuntimeFeatures(output); err != nil {
 			logrus.Errorf("Unable to load OCI features for runtime handler %q: %v", name, err)
+
 			continue
 		}
 
@@ -1393,9 +1527,11 @@ func (c *RuntimeConfig) initializeRuntimeFeatures() {
 			// but the current kernel might not.
 			if err := checkKernelRROMountSupport(); err != nil {
 				logrus.Warnf("Runtime handler %q supports Recursive Read-only (RRO) mounts, but kernel does not: %v", name, err)
+
 				rro = false
 			}
 		}
+
 		handler.features.RecursiveReadOnlyMounts = rro
 	}
 }
@@ -1408,6 +1544,7 @@ func (c *RuntimeConfig) TranslateMonitorFields(onExecution bool) error {
 			}
 		}
 	}
+
 	return nil
 }
 
@@ -1418,31 +1555,45 @@ func (c *RuntimeConfig) TranslateMonitorFieldsForHandler(handler *RuntimeHandler
 		logrus.Debugf("Monitor cgroup %s is becoming %s", handler.MonitorCgroup, c.ConmonCgroup)
 		handler.MonitorCgroup = c.ConmonCgroup
 	}
+
 	if c.Conmon != "" {
 		logrus.Debugf("Monitor path %s is becoming %s", handler.MonitorPath, c.Conmon)
 		handler.MonitorPath = c.Conmon
 	}
+
 	if len(c.ConmonEnv) != 0 {
 		handler.MonitorEnv = c.ConmonEnv
 	}
-	// If empty, assume default
-	if handler.MonitorCgroup == "" {
+
+	// If systemd and empty, assume default
+	if c.cgroupManager.IsSystemd() && handler.MonitorCgroup == "" {
 		handler.MonitorCgroup = defaultMonitorCgroup
 	}
+
 	if onExecution {
 		if err := c.ValidateConmonPath("conmon", handler); err != nil {
 			return err
 		}
+		// if cgroupManager is cgroupfs
 		if !c.cgroupManager.IsSystemd() {
-			if handler.MonitorCgroup != utils.PodCgroupName && handler.MonitorCgroup != "" {
-				return errors.New("cgroupfs manager conmon cgroup should be 'pod' or empty")
+			// handler.MonitorCgroup having value "" is valid
+			// but the default value system.slice is not
+			if handler.MonitorCgroup == defaultMonitorCgroup {
+				handler.MonitorCgroup = ""
 			}
+
+			if handler.MonitorCgroup != utils.PodCgroupName && handler.MonitorCgroup != "" {
+				return fmt.Errorf("cgroupfs manager conmon cgroup should be 'pod' or empty, but got: '%s'", handler.MonitorCgroup)
+			}
+
 			return nil
 		}
-		if !(handler.MonitorCgroup == utils.PodCgroupName || strings.HasSuffix(handler.MonitorCgroup, ".slice")) {
+
+		if handler.MonitorCgroup != utils.PodCgroupName && !strings.HasSuffix(handler.MonitorCgroup, ".slice") {
 			return errors.New("conmon cgroup should be 'pod' or a systemd slice")
 		}
 	}
+
 	return nil
 }
 
@@ -1451,10 +1602,12 @@ func (c *RuntimeConfig) TranslateMonitorFieldsForHandler(handler *RuntimeHandler
 // In any other case, it simply checks if `Conmon` is a valid file.
 func (c *RuntimeConfig) ValidateConmonPath(executable string, handler *RuntimeHandler) error {
 	var err error
+
 	handler.MonitorPath, err = validateExecutablePath(executable, handler.MonitorPath)
 	if err != nil {
 		return err
 	}
+
 	c.conmonManager, err = conmonmgr.New(handler.MonitorPath)
 
 	return err
@@ -1523,13 +1676,18 @@ func validateExecutablePath(executable, currentPath string) (string, error) {
 		if err != nil {
 			return "", err
 		}
+
 		logrus.Debugf("Using %s from $PATH: %s", executable, path)
+
 		return path, nil
 	}
+
 	if _, err := os.Stat(currentPath); err != nil {
 		return "", fmt.Errorf("invalid %s path: %w", executable, err)
 	}
+
 	logrus.Infof("Using %s executable: %s", executable, currentPath)
+
 	return currentPath, nil
 }
 
@@ -1539,14 +1697,23 @@ func (c *ImageConfig) Validate(onExecution bool) error {
 	if !filepath.IsAbs(c.SignaturePolicyDir) {
 		return fmt.Errorf("signature policy dir %q is not absolute", c.SignaturePolicyDir)
 	}
+
 	if _, err := c.ParsePauseImage(); err != nil {
 		return fmt.Errorf("invalid pause image %q: %w", c.PauseImage, err)
 	}
+
+	switch c.ShortNameMode {
+	case "enforcing", "disabled", "":
+	default:
+		return fmt.Errorf("invalid short name mode %q", c.ShortNameMode)
+	}
+
 	if onExecution {
 		if err := os.MkdirAll(c.SignaturePolicyDir, 0o755); err != nil {
 			return fmt.Errorf("cannot create signature policy dir: %w", err)
 		}
 	}
+
 	return nil
 }
 
@@ -1580,6 +1747,7 @@ func (c *NetworkConfig) Validate(onExecution bool) error {
 		// While the plugin_dir option is being deprecated, we need this check
 		if c.PluginDir != "" {
 			logrus.Warnf("The config field plugin_dir is being deprecated. Please use plugin_dirs instead")
+
 			if err := os.MkdirAll(c.PluginDir, 0o755); err != nil {
 				return fmt.Errorf("invalid plugin_dir entry: %w", err)
 			}
@@ -1599,6 +1767,7 @@ func (c *NetworkConfig) Validate(onExecution bool) error {
 		if err != nil {
 			return fmt.Errorf("initialize CNI plugin: %w", err)
 		}
+
 		c.cniManager = cniManager
 	}
 
@@ -1610,20 +1779,40 @@ func (r *RuntimeHandler) Validate(name string) error {
 	if err := r.ValidateRuntimeType(name); err != nil {
 		return err
 	}
+
 	if err := r.ValidateRuntimePath(name); err != nil {
 		return err
 	}
+
 	if err := r.ValidateRuntimeConfigPath(name); err != nil {
 		return err
 	}
+
 	if err := r.ValidateRuntimeAllowedAnnotations(); err != nil {
 		return err
 	}
+
 	if err := r.ValidateContainerMinMemory(name); err != nil {
 		logrus.Errorf("Unable to set minimum container memory for runtime handler %q: %v", name, err)
 	}
 
-	return r.ValidateNoSyncLog()
+	if err := r.ValidateNoSyncLog(); err != nil {
+		return fmt.Errorf("no sync log: %w", err)
+	}
+
+	if err := r.ValidateWebsocketStreaming(name); err != nil {
+		return fmt.Errorf("websocket streaming: %w", err)
+	}
+
+	if err := r.validateRuntimeExecCPUAffinity(); err != nil {
+		return err
+	}
+
+	if err := r.validateRuntimeSeccompProfile(); err != nil {
+		return err
+	}
+
+	return nil
 }
 
 func (r *RuntimeHandler) ValidateRuntimeVMBinaryPattern() bool {
@@ -1650,6 +1839,7 @@ func (r *RuntimeHandler) ValidateRuntimePath(name string) error {
 		if err != nil {
 			return fmt.Errorf("%q not found in $PATH: %w", name, err)
 		}
+
 		r.RuntimePath = executable
 		logrus.Debugf("Using runtime executable from $PATH %q", executable)
 	} else if _, err := os.Stat(r.RuntimePath); err != nil && os.IsNotExist(err) {
@@ -1665,6 +1855,7 @@ func (r *RuntimeHandler) ValidateRuntimePath(name string) error {
 	logrus.Debugf(
 		"Found valid runtime %q for runtime_path %q", name, r.RuntimePath,
 	)
+
 	return nil
 }
 
@@ -1674,6 +1865,7 @@ func (r *RuntimeHandler) ValidateRuntimeType(name string) error {
 		return fmt.Errorf("invalid `runtime_type` %q for runtime %q",
 			r.RuntimeType, name)
 	}
+
 	return nil
 }
 
@@ -1682,12 +1874,15 @@ func (r *RuntimeHandler) ValidateRuntimeConfigPath(name string) error {
 	if r.RuntimeConfigPath == "" {
 		return nil
 	}
+
 	if r.RuntimeType != RuntimeTypeVM {
 		return errors.New("runtime_config_path can only be used with the 'vm' runtime type")
 	}
+
 	if _, err := os.Stat(r.RuntimeConfigPath); err != nil && os.IsNotExist(err) {
 		return fmt.Errorf("invalid runtime_config_path for runtime '%s': %w", name, err)
 	}
+
 	return nil
 }
 
@@ -1696,10 +1891,13 @@ func (r *RuntimeHandler) ValidateRuntimeAllowedAnnotations() error {
 	if err != nil {
 		return err
 	}
+
 	logrus.Debugf(
 		"Allowed annotations for runtime: %v", r.AllowedAnnotations,
 	)
+
 	r.DisallowedAnnotations = disallowed
+
 	return nil
 }
 
@@ -1712,8 +1910,10 @@ func (r *RuntimeHandler) ValidateNoSyncLog() error {
 	// This means that the runtime type must be set to 'oci' or left empty
 	if r.RuntimeType == DefaultRuntimeType || r.RuntimeType == "" {
 		logrus.Warn("NoSyncLog is enabled. This can lead to lost log data")
+
 		return nil
 	}
+
 	return fmt.Errorf("no_sync_log is only allowed with runtime type 'oci', runtime type is '%s'", r.RuntimeType)
 }
 
@@ -1729,9 +1929,53 @@ func (r *RuntimeHandler) ValidateContainerMinMemory(name string) error {
 		err = fmt.Errorf("unable to set runtime memory to %q: %w. Setting to %q instead", r.ContainerMinMemory, err, defaultContainerMinMemory)
 		// Fallback to default value if something is wrong with the configured value.
 		r.ContainerMinMemory = units.BytesSize(defaultContainerMinMemory)
+
 		return err
 	}
+
 	logrus.Debugf("Runtime handler %q container minimum memory set to %d bytes", name, memorySize)
+
+	return nil
+}
+
+// ValidateWebsocketStreaming can be used to verify if the runtime supports WebSocket streaming.
+func (r *RuntimeHandler) ValidateWebsocketStreaming(name string) error {
+	if r.RuntimeType != RuntimeTypePod {
+		if r.StreamWebsockets {
+			return fmt.Errorf(`only the 'runtime_type = "pod"' supports websocket streaming, not %q (runtime %q)`, r.RuntimeType, name)
+		}
+
+		return nil
+	}
+
+	// Requires at least conmon-rs v0.7.0
+	v, err := conmonrsClient.Version(r.MonitorPath)
+	if err != nil {
+		if errors.Is(err, conmonrsClient.ErrUnsupported) {
+			logrus.Debugf("Unable to verify pod runtime version: %v", err)
+
+			// Streaming server support got introduced in v0.7.0
+			if r.StreamWebsockets {
+				logrus.Warnf("Disabling streaming over websockets, it requires conmon-rs >= v0.7.0")
+
+				r.StreamWebsockets = false
+			}
+
+			return nil
+		}
+
+		return fmt.Errorf("get conmon-rs version: %w", err)
+	}
+
+	if v.Tag == "" {
+		v.Tag = "none"
+	}
+
+	logrus.Infof(
+		"Runtime handler %q is using conmon-rs version: %s, tag: %s, commit: %s, build: %s, target: %s, %s, %s",
+		name, v.Version, v.Tag, v.Commit, v.BuildDate, v.Target, v.RustVersion, v.CargoVersion,
+	)
+
 	return nil
 }
 
@@ -1764,9 +2008,11 @@ func (r *RuntimeHandler) RuntimeSupportsIDMap() bool {
 	if r.features.Linux == nil || r.features.Linux.MountExtensions == nil || r.features.Linux.MountExtensions.IDMap == nil {
 		return false
 	}
+
 	if enabled := r.features.Linux.MountExtensions.IDMap.Enabled; enabled == nil || !*enabled {
 		return false
 	}
+
 	return true
 }
 
@@ -1785,23 +2031,64 @@ func (r *RuntimeHandler) RuntimeDefaultAnnotations() map[string]string {
 	return r.DefaultAnnotations
 }
 
+// RuntimeStreamWebsockets returns the configured websocket streaming option for this handler.
+func (r *RuntimeHandler) RuntimeStreamWebsockets() bool {
+	return r.StreamWebsockets
+}
+
+// RuntimeSeccomp returns the configuration of the loaded seccomp profile for this handler.
+func (r *RuntimeHandler) RuntimeSeccomp() *seccomp.Config {
+	return r.seccompConfig
+}
+
+// validateRuntimeExecCPUAffinity checks if the RuntimeHandler enforces proper CPU affinity settings.
+func (r *RuntimeHandler) validateRuntimeExecCPUAffinity() error {
+	switch r.ExecCPUAffinity {
+	case ExecCPUAffinityTypeDefault, ExecCPUAffinityTypeFirst:
+		return nil
+	}
+
+	return fmt.Errorf("invalid exec_cpu_affinity %q", r.ExecCPUAffinity)
+}
+
+// validateRuntimeSeccompProfile tries to load the RuntimeHandler seccomp profile.
+func (r *RuntimeHandler) validateRuntimeSeccompProfile() error {
+	if r.SeccompProfile == "" {
+		r.seccompConfig = nil
+
+		return nil
+	}
+
+	r.seccompConfig = seccomp.New()
+	if err := r.seccompConfig.LoadProfile(r.SeccompProfile); err != nil {
+		return fmt.Errorf("unable to load runtime handler seccomp profile: %w", err)
+	}
+
+	return nil
+}
+
 func validateAllowedAndGenerateDisallowedAnnotations(allowed []string) (disallowed []string, _ error) {
 	disallowedMap := make(map[string]bool)
 	for _, ann := range annotations.AllAllowedAnnotations {
 		disallowedMap[ann] = false
 	}
+
 	for _, ann := range allowed {
 		if _, ok := disallowedMap[ann]; !ok {
 			return nil, fmt.Errorf("invalid allowed_annotation: %s", ann)
 		}
+
 		disallowedMap[ann] = true
 	}
+
 	disallowed = make([]string, 0, len(disallowedMap))
+
 	for ann, allowed := range disallowedMap {
 		if !allowed {
 			disallowed = append(disallowed, ann)
 		}
 	}
+
 	return disallowed, nil
 }
 
@@ -1820,6 +2107,14 @@ func (c *NetworkConfig) CNIPluginAddWatcher() chan bool {
 	return c.cniManager.AddWatcher()
 }
 
+// CNIPluginGC calls the plugin's GC to clean up any resources concerned with
+// stale pods (pod other than the ones provided by validPodList). The call to
+// the plugin will be deferred until it is ready logging any errors then and
+// returning nil error here.
+func (c *Config) CNIPluginGC(ctx context.Context, validPodList cnimgr.PodNetworkLister) error {
+	return c.cniManager.GC(ctx, validPodList)
+}
+
 // CNIManagerShutdown shuts down the CNI Manager.
 func (c *NetworkConfig) CNIManagerShutdown() {
 	c.cniManager.Shutdown()
