@@ -11,6 +11,7 @@ import (
 	. "github.com/onsi/ginkgo/v2"
 	. "github.com/onsi/gomega"
 
+	"github.com/cri-o/cri-o/internal/config/cgmgr"
 	crioann "github.com/cri-o/cri-o/pkg/annotations"
 	"github.com/cri-o/cri-o/pkg/config"
 	"github.com/cri-o/cri-o/utils/cmdrunner"
@@ -33,11 +34,12 @@ var _ = t.Describe("Config", func() {
 		sut.NamespacesDir = os.TempDir()
 		sut.Conmon = validConmonPath()
 		tmpDir := t.MustTempDir("cni-test")
-		sut.NetworkConfig.PluginDirs = []string{tmpDir}
+		sut.PluginDirs = []string{tmpDir}
 		sut.NetworkDir = os.TempDir()
 		sut.LogDir = "/"
 		sut.Listen = t.MustTempFile("crio.sock")
 		sut.HooksDir = []string{}
+
 		return sut
 	}
 
@@ -73,7 +75,7 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail with invalid log_dir", func() {
 			// Given
-			sut.RootConfig.LogDir = "/dev/null"
+			sut.LogDir = "/dev/null"
 
 			// When
 			err := sut.Validate(true)
@@ -109,7 +111,7 @@ var _ = t.Describe("Config", func() {
 			// Given
 			sut.Runtimes[config.DefaultRuntime] = &config.RuntimeHandler{RuntimePath: validDirPath}
 			sut.Conmon = validConmonPath()
-			sut.NetworkConfig.NetworkDir = invalidPath
+			sut.NetworkDir = invalidPath
 
 			// When
 			err := sut.Validate(true)
@@ -187,6 +189,51 @@ var _ = t.Describe("Config", func() {
 			// Then
 			Expect(err).To(HaveOccurred())
 		})
+
+		It("should succeed if stream server TLS enabled", func() {
+			// Given
+			sut = runtimeValidConfig()
+			sut.StreamEnableTLS = true
+			sut.StreamTLSCert = "cert"
+			sut.StreamTLSKey = "key"
+			sut.StreamTLSCA = "ca"
+
+			// When
+			err := sut.APIConfig.Validate(false)
+
+			// Then
+			Expect(err).NotTo(HaveOccurred())
+		})
+
+		It("should fail if stream server TLS enabled and cert is empty", func() {
+			// Given
+			sut = runtimeValidConfig()
+			sut.StreamEnableTLS = true
+			sut.StreamTLSCert = ""
+			sut.StreamTLSKey = "key"
+			sut.StreamTLSCA = "ca"
+
+			// When
+			err := sut.APIConfig.Validate(false)
+
+			// Then
+			Expect(err).To(HaveOccurred())
+		})
+
+		It("should fail if stream server TLS enabled and key is empty", func() {
+			// Given
+			sut = runtimeValidConfig()
+			sut.StreamEnableTLS = true
+			sut.StreamTLSCert = "cert"
+			sut.StreamTLSKey = ""
+			sut.StreamTLSCA = "ca"
+
+			// When
+			err := sut.APIConfig.Validate(false)
+
+			// Then
+			Expect(err).To(HaveOccurred())
+		})
 	})
 
 	t.Describe("ValidateRuntimeConfig", func() {
@@ -472,7 +519,7 @@ var _ = t.Describe("Config", func() {
 			sut.ConmonCgroup = "wrong"
 
 			// When
-			err := sut.RuntimeConfig.TranslateMonitorFieldsForHandler(handler, true)
+			err := sut.TranslateMonitorFieldsForHandler(handler, true)
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -484,15 +531,29 @@ var _ = t.Describe("Config", func() {
 			sut.ConmonCgroup = invalid
 
 			// When
-			err := sut.RuntimeConfig.TranslateMonitorFieldsForHandler(handler, true)
+			err := sut.TranslateMonitorFieldsForHandler(handler, true)
 
 			// Then
 			Expect(err).To(HaveOccurred())
 		})
+		It("should not fail on cgroupfs as cgroup manager and conmon_cgroup as empty", func() {
+			handler := &config.RuntimeHandler{}
+
+			// Given
+			cgm, _ := cgmgr.SetCgroupManager("cgroupfs") //nolint:errcheck
+			runtimeConfig := *config.DefaultRuntimeConfig(cgm)
+
+			// When
+			err := runtimeConfig.TranslateMonitorFieldsForHandler(handler, true)
+
+			// Then
+			Expect(handler.MonitorCgroup).To(Equal(""))
+			Expect(err).ToNot(HaveOccurred())
+		})
 
 		It("should fail on invalid InfraCtrCPUSet", func() {
 			// Given
-			sut.RuntimeConfig.InfraCtrCPUSet = "unparsable"
+			sut.InfraCtrCPUSet = "unparsable"
 
 			// When
 			err := sut.RuntimeConfig.Validate(nil, false)
@@ -507,7 +568,7 @@ var _ = t.Describe("Config", func() {
 			handler := &config.RuntimeHandler{}
 
 			// When
-			err := sut.RuntimeConfig.TranslateMonitorFieldsForHandler(handler, true)
+			err := sut.TranslateMonitorFieldsForHandler(handler, true)
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -518,7 +579,7 @@ var _ = t.Describe("Config", func() {
 			handler := &config.RuntimeHandler{}
 
 			// When
-			err := sut.RuntimeConfig.TranslateMonitorFieldsForHandler(handler, true)
+			err := sut.TranslateMonitorFieldsForHandler(handler, true)
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -526,11 +587,11 @@ var _ = t.Describe("Config", func() {
 		})
 		It("should inherit from .ConmonEnv", func() {
 			// Given
-			sut.ConmonEnv = []string{"PATH=/usr/bin"}
+			sut.ConmonEnv = []string{"PATH=/foo/bar/baz"}
 			handler := &config.RuntimeHandler{}
 
 			// When
-			err := sut.RuntimeConfig.TranslateMonitorFieldsForHandler(handler, true)
+			err := sut.TranslateMonitorFieldsForHandler(handler, false)
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -542,7 +603,7 @@ var _ = t.Describe("Config", func() {
 			handler := &config.RuntimeHandler{}
 
 			// When
-			err := sut.RuntimeConfig.TranslateMonitorFieldsForHandler(handler, true)
+			err := sut.TranslateMonitorFieldsForHandler(handler, false)
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -557,7 +618,7 @@ var _ = t.Describe("Config", func() {
 
 			// Given
 			cmdrunner.ResetPrependedCmd()
-			sut.RuntimeConfig.InfraCtrCPUSet = "0"
+			sut.InfraCtrCPUSet = "0"
 
 			// When
 			err = sut.RuntimeConfig.Validate(nil, false)
@@ -570,7 +631,7 @@ var _ = t.Describe("Config", func() {
 		It("should not configure a taskset prefix for cmdrunner for an empty InfraCtrCPUSet", func() {
 			// Given
 			cmdrunner.ResetPrependedCmd()
-			sut.RuntimeConfig.InfraCtrCPUSet = ""
+			sut.InfraCtrCPUSet = ""
 
 			// When
 			err := sut.RuntimeConfig.Validate(nil, false)
@@ -585,7 +646,7 @@ var _ = t.Describe("Config", func() {
 		It("should succeed with default config", func() {
 			// Given
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -598,7 +659,7 @@ var _ = t.Describe("Config", func() {
 			}
 
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -610,7 +671,7 @@ var _ = t.Describe("Config", func() {
 			sut.DefaultRuntime = invalidPath
 
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -622,7 +683,7 @@ var _ = t.Describe("Config", func() {
 			sut.DefaultRuntime = config.DefaultRuntime
 
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -633,7 +694,7 @@ var _ = t.Describe("Config", func() {
 			sut.Runtimes[config.DefaultRuntime] = &config.RuntimeHandler{RuntimePath: invalidPath}
 
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -647,7 +708,7 @@ var _ = t.Describe("Config", func() {
 			}
 
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -661,7 +722,7 @@ var _ = t.Describe("Config", func() {
 			}
 
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -674,7 +735,7 @@ var _ = t.Describe("Config", func() {
 			}
 
 			// When
-			err := sut.RuntimeConfig.ValidateRuntimes()
+			err := sut.ValidateRuntimes()
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -718,16 +779,105 @@ var _ = t.Describe("Config", func() {
 			Expect(err).To(HaveOccurred())
 			Expect(err).To(MatchError("no_sync_log is only allowed with runtime type 'oci', runtime type is 'vm'"))
 		})
+
+		It("should disallow stream_websockets for the 'oci' runtime", func() {
+			sut.Runtimes[config.DefaultRuntime] = &config.RuntimeHandler{
+				RuntimePath:      validFilePath,
+				RuntimeType:      config.DefaultRuntimeType,
+				StreamWebsockets: true,
+			}
+
+			err := sut.Runtimes[config.DefaultRuntime].ValidateWebsocketStreaming(config.DefaultRuntime)
+
+			Expect(err).To(HaveOccurred())
+			Expect(err).To(MatchError(`only the 'runtime_type = "pod"' supports websocket streaming, not "oci" (runtime "crun")`))
+		})
+
+		It("should allow 'stream_websockets == false' for the 'oci' runtime", func() {
+			sut.Runtimes[config.DefaultRuntime] = &config.RuntimeHandler{
+				RuntimePath:      validFilePath,
+				RuntimeType:      config.DefaultRuntimeType,
+				StreamWebsockets: false,
+			}
+
+			err := sut.Runtimes[config.DefaultRuntime].ValidateWebsocketStreaming(config.DefaultRuntime)
+
+			Expect(err).NotTo(HaveOccurred())
+		})
+
+		conmonrsFakeBinary := func(command string) (name string) {
+			file, err := os.CreateTemp("", "conmonrs-fake-*")
+			Expect(err).NotTo(HaveOccurred())
+
+			err = file.Chmod(0o755)
+			Expect(err).NotTo(HaveOccurred())
+
+			_, err = file.WriteString("#!/bin/sh\n" + command)
+			Expect(err).NotTo(HaveOccurred())
+
+			err = file.Close()
+			Expect(err).NotTo(HaveOccurred())
+
+			return file.Name()
+		}
+
+		It("should support streaming websockets if conmon-rs is >= v0.7.0", func() {
+			fileName := conmonrsFakeBinary("echo '{ \"version\": \"0.7.0\" }'")
+			defer os.RemoveAll(fileName)
+
+			sut.Runtimes[config.DefaultRuntime] = &config.RuntimeHandler{
+				RuntimePath:      validFilePath,
+				RuntimeType:      config.RuntimeTypePod,
+				StreamWebsockets: true,
+				MonitorPath:      fileName,
+			}
+
+			err := sut.Runtimes[config.DefaultRuntime].ValidateWebsocketStreaming(config.DefaultRuntime)
+			Expect(err).NotTo(HaveOccurred())
+			Expect(sut.Runtimes[config.DefaultRuntime].StreamWebsockets).To(BeTrue())
+		})
+
+		It("should disable streaming websockets if conmon-rs is < v0.7.0", func() {
+			fileName := conmonrsFakeBinary("echo 'error: unexpected argument' && exit 1")
+			defer os.RemoveAll(fileName)
+
+			sut.Runtimes[config.DefaultRuntime] = &config.RuntimeHandler{
+				RuntimePath:      validFilePath,
+				RuntimeType:      config.RuntimeTypePod,
+				StreamWebsockets: true,
+				MonitorPath:      fileName,
+			}
+
+			err := sut.Runtimes[config.DefaultRuntime].ValidateWebsocketStreaming(config.DefaultRuntime)
+			Expect(err).NotTo(HaveOccurred())
+			Expect(sut.Runtimes[config.DefaultRuntime].StreamWebsockets).To(BeFalse())
+		})
+
+		It("should not disable streaming websockets if conmon-rs is version is not retrievable", func() {
+			fileName := conmonrsFakeBinary("exit 1")
+			defer os.RemoveAll(fileName)
+
+			sut.Runtimes[config.DefaultRuntime] = &config.RuntimeHandler{
+				RuntimePath:      validFilePath,
+				RuntimeType:      config.RuntimeTypePod,
+				StreamWebsockets: true,
+				MonitorPath:      fileName,
+			}
+
+			err := sut.Runtimes[config.DefaultRuntime].ValidateWebsocketStreaming(config.DefaultRuntime)
+			Expect(err).To(HaveOccurred())
+			Expect(sut.Runtimes[config.DefaultRuntime].StreamWebsockets).To(BeTrue())
+		})
 	})
 
 	t.Describe("ValidateConmonPath", func() {
 		It("should succeed with valid file in $PATH", func() {
 			// Given
-			sut.RuntimeConfig.Conmon = ""
+			sut.Conmon = ""
 			handler := &config.RuntimeHandler{MonitorPath: ""}
 
 			// When
-			err := sut.RuntimeConfig.ValidateConmonPath(validConmonPath(), handler)
+			err := sut.ValidateConmonPath(validConmonPath(), handler)
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -739,7 +889,7 @@ var _ = t.Describe("Config", func() {
 			handler := &config.RuntimeHandler{MonitorPath: ""}
 
 			// When
-			err := sut.RuntimeConfig.ValidateConmonPath(invalidPath, handler)
+			err := sut.ValidateConmonPath(invalidPath, handler)
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -750,7 +900,7 @@ var _ = t.Describe("Config", func() {
 			handler := &config.RuntimeHandler{MonitorPath: validConmonPath()}
 
 			// When
-			err := sut.RuntimeConfig.ValidateConmonPath("", handler)
+			err := sut.ValidateConmonPath("", handler)
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -761,7 +911,7 @@ var _ = t.Describe("Config", func() {
 			handler := &config.RuntimeHandler{MonitorPath: invalidPath}
 
 			// When
-			err := sut.RuntimeConfig.ValidateConmonPath("", handler)
+			err := sut.ValidateConmonPath("", handler)
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -780,7 +930,7 @@ var _ = t.Describe("Config", func() {
 
 		It("should succeed on execution and writing permissions", func() {
 			// Given
-			sut.ImageConfig.SignaturePolicyDir = os.TempDir()
+			sut.SignaturePolicyDir = os.TempDir()
 
 			// When
 			err := sut.ImageConfig.Validate(true)
@@ -791,7 +941,7 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail when SignaturePolicyDir is not absolute", func() {
 			// Given
-			sut.ImageConfig.SignaturePolicyDir = "./wrong/path"
+			sut.SignaturePolicyDir = "./wrong/path"
 
 			// When
 			err := sut.ImageConfig.Validate(false)
@@ -802,7 +952,7 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail when PauseImage is invalid", func() {
 			// Given
-			sut.ImageConfig.PauseImage = "//NOT:a valid image reference!"
+			sut.PauseImage = "//NOT:a valid image reference!"
 
 			// When
 			err := sut.ImageConfig.Validate(false)
@@ -815,10 +965,10 @@ var _ = t.Describe("Config", func() {
 	t.Describe("ImageConfig.ParsePauseImage", func() {
 		It("should succeed with the default value", func() {
 			// Given
-			sut.ImageConfig.PauseImage = config.DefaultPauseImage
+			sut.PauseImage = config.DefaultPauseImage
 
 			// When
-			ref, err := sut.ImageConfig.ParsePauseImage()
+			ref, err := sut.ParsePauseImage()
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -828,10 +978,10 @@ var _ = t.Describe("Config", func() {
 
 		It("should succeed with a name-only value", func() {
 			// Given
-			sut.ImageConfig.PauseImage = "registry.k8s.io/pause"
+			sut.PauseImage = "registry.k8s.io/pause"
 
 			// When
-			ref, err := sut.ImageConfig.ParsePauseImage()
+			ref, err := sut.ParsePauseImage()
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -843,10 +993,10 @@ var _ = t.Describe("Config", func() {
 			// name with a registry
 
 			// Given
-			sut.ImageConfig.PauseImage = "short:notlatest"
+			sut.PauseImage = "short:notlatest"
 
 			// When
-			ref, err := sut.ImageConfig.ParsePauseImage()
+			ref, err := sut.ParsePauseImage()
 
 			// Then
 			Expect(err).ToNot(HaveOccurred())
@@ -855,10 +1005,10 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail with an invalid value", func() {
 			// Given
-			sut.ImageConfig.PauseImage = "//THIS is:very!invalid="
+			sut.PauseImage = "//THIS is:very!invalid="
 
 			// When
-			_, err := sut.ImageConfig.ParsePauseImage()
+			_, err := sut.ParsePauseImage()
 
 			// Then
 			Expect(err).To(HaveOccurred())
@@ -891,8 +1041,8 @@ var _ = t.Describe("Config", func() {
 			tmpDir := t.MustTempDir("network")
 			Expect(os.RemoveAll(tmpDir)).ToNot(HaveOccurred())
 
-			sut.NetworkConfig.NetworkDir = tmpDir
-			sut.NetworkConfig.PluginDirs = []string{validDirPath}
+			sut.NetworkDir = tmpDir
+			sut.PluginDirs = []string{validDirPath}
 
 			// When
 			err := sut.NetworkConfig.Validate(true)
@@ -908,8 +1058,8 @@ var _ = t.Describe("Config", func() {
 			Expect(err).ToNot(HaveOccurred())
 			file.Close()
 			defer os.Remove(tmpfile)
-			sut.NetworkConfig.NetworkDir = tmpfile
-			sut.NetworkConfig.PluginDirs = []string{}
+			sut.NetworkDir = tmpfile
+			sut.PluginDirs = []string{}
 
 			// When
 			err = sut.NetworkConfig.Validate(true)
@@ -920,8 +1070,8 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail on invalid PluginDirs", func() {
 			// Given
-			sut.NetworkConfig.NetworkDir = validDirPath
-			sut.NetworkConfig.PluginDirs = []string{invalidPath}
+			sut.NetworkDir = validDirPath
+			sut.PluginDirs = []string{invalidPath}
 
 			// When
 			err := sut.NetworkConfig.Validate(true)
@@ -932,9 +1082,9 @@ var _ = t.Describe("Config", func() {
 
 		It("should succeed on having PluginDir", func() {
 			// Given
-			sut.NetworkConfig.NetworkDir = validDirPath
-			sut.NetworkConfig.PluginDir = validDirPath
-			sut.NetworkConfig.PluginDirs = []string{}
+			sut.NetworkDir = validDirPath
+			sut.PluginDir = validDirPath
+			sut.PluginDirs = []string{}
 
 			// When
 			err := sut.NetworkConfig.Validate(true)
@@ -945,9 +1095,9 @@ var _ = t.Describe("Config", func() {
 
 		It("should succeed in appending PluginDir to PluginDirs", func() {
 			// Given
-			sut.NetworkConfig.NetworkDir = validDirPath
-			sut.NetworkConfig.PluginDir = validDirPath
-			sut.NetworkConfig.PluginDirs = []string{}
+			sut.NetworkDir = validDirPath
+			sut.PluginDir = validDirPath
+			sut.PluginDirs = []string{}
 
 			// When
 			err := sut.NetworkConfig.Validate(true)
@@ -959,9 +1109,9 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail in validating invalid PluginDir", func() {
 			// Given
-			sut.NetworkConfig.NetworkDir = validDirPath
-			sut.NetworkConfig.PluginDir = invalidPath
-			sut.NetworkConfig.PluginDirs = []string{}
+			sut.NetworkDir = validDirPath
+			sut.PluginDir = invalidPath
+			sut.PluginDirs = []string{}
 
 			// When
 			err := sut.NetworkConfig.Validate(true)
@@ -998,7 +1148,7 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail on invalid LogDir", func() {
 			// Given
-			sut.RootConfig.LogDir = "/dev/null"
+			sut.LogDir = "/dev/null"
 
 			// When
 			err := sut.RootConfig.Validate(true)
@@ -1009,7 +1159,7 @@ var _ = t.Describe("Config", func() {
 
 		It("should fail with non absolute log_dir", func() {
 			// Given
-			sut.RootConfig.LogDir = "test"
+			sut.LogDir = "test"
 
 			// When
 			err := sut.Validate(true)
@@ -1027,12 +1177,12 @@ var _ = t.Describe("Config", func() {
 			defaultStore, err := storage.GetStore(storage.StoreOptions{})
 			Expect(err).ToNot(HaveOccurred())
 
-			sut.RootConfig.RunRoot = ""
-			sut.RootConfig.Root = ""
-			sut.RootConfig.Storage = ""
-			sut.RootConfig.StorageOptions = make([]string, 0)
+			sut.RunRoot = ""
+			sut.Root = ""
+			sut.Storage = ""
+			sut.StorageOptions = make([]string, 0)
 			// this must be set in case pinns isn't downloaded to the $PATH
-			sut.RuntimeConfig.PinnsPath = alwaysPresentPath
+			sut.PinnsPath = alwaysPresentPath
 
 			// When
 			err = sut.Validate(true)
@@ -1054,10 +1204,10 @@ var _ = t.Describe("Config", func() {
 			defaultStore, err := storage.GetStore(storage.StoreOptions{})
 			Expect(err).ToNot(HaveOccurred())
 
-			sut.RootConfig.RunRoot = alwaysPresentPath
-			sut.RootConfig.Root = alwaysPresentPath
+			sut.RunRoot = alwaysPresentPath
+			sut.Root = alwaysPresentPath
 			// this must be set in case pinns isn't downloaded to the $PATH
-			sut.RuntimeConfig.PinnsPath = alwaysPresentPath
+			sut.PinnsPath = alwaysPresentPath
 
 			// When
 			err = sut.Validate(true)
